#ifndef CUDADEVICE_H
#define CUDADEVICE_H

#include <cuda.h>
#include "xmlreader.h"

#define CUDA_CHAR_BUFFER_SIZE 1024

#define __ALIGN(PAR) offset = (offset + __alignof(PAR) - 1) & ~(__alignof(PAR) - 1);

#define REGS__prepareCellPos    11
#define REGS__prepareCellMate   16
#define REGS__prepareCellMask   7

#define REGS__StabilisationBase 16

#define maxStabilLoops 1000

class CudaDevice
{
  bool          isInit, isContext;
  int           _id, _cMaj, _cMin, attribute[40];
  size_t        _totalMem, _infoMemTotal, _infoMemFree;
  QString       _name;

  CUdevice      device;
  CUdevprop     props;
  CUcontext     context;
  CUmodule      module;

  CUdeviceptr   posDevPtr;

public:
  CudaDevice(int devId = 0);

  inline bool initialized(){ return isInit; }

  inline int id() { return _id; }
  inline int cMaj() { return _cMaj; }
  inline int cMin() { return _cMin; }

  inline size_t totalMem() { return _totalMem; }
  inline size_t infoMemTotal() { return _infoMemTotal; }
  inline size_t infoMemFree() { return _infoMemFree; }

  inline QString name() { return _name; }

  inline bool isDoublePrecission() { if(_cMaj > 1)return true; if(_cMaj < 1)return false; return _cMin == 3; }

  bool createContext();
  bool detachContext();

  bool loadModule(QString fileName);
  bool freeModule();

  bool memAlloc(CUdeviceptr * ptr, size_t memSize);
  bool memFree(CUdeviceptr ptr);
  bool memcpyHtoD(CUdeviceptr ptr, void * src, size_t memSize);
  bool memcpyDtoH(void * out, CUdeviceptr ptr, size_t memSize);

  bool moduleGetFunction(const char * funName, CUfunction * func);
  bool paramSetSize(CUfunction func, int offset);
  bool funcSetBlockShape(CUfunction func, int x, int y, int z);
  bool funcSetSharedSize(CUfunction func, unsigned int numBytes);

  bool launchGrid(CUfunction func, bool useGrid, int width, int height);

  bool paramSetI( CUfunction func, int * offset, size_t addOffset, int val, int * storeOffset = 0);
  bool paramSetUL(CUfunction func, int * offset, size_t addOffset, unsigned long val, int * storeOffset = 0);
  bool paramSetF( CUfunction func, int * offset, size_t addOffset, float val, int * storeOffset = 0);
  bool paramSetD( CUfunction func, int * offset, size_t addOffset, double val, int * storeOffset = 0);
  bool paramSetV( CUfunction func, int * offset, size_t addOffset, void * ptr, unsigned int numBytes, int * storeOffset = 0);
/*
  bool allocPos(VecF * data, int count);
  bool freePos();
*/
  bool prepareCellPos(ProjectF & prj, VecF * allPositions, VecF & begin, int stepAlpha, int stepBeta, int stepGama);
  bool prepareCellMate(ProjectF & prj, CellMate * allMates, int stepAlpha, int stepBeta, int stepGama);
  bool prepareCellMask(ProjectF & prj, VecF * allPositions, ulong * allMask, int stepAlpha, int stepBeta, int stepGama);

  bool prepareMateOffset(CrystalMatrixType type, int matrix, int * mateID, int stepAlpha, int stepBeta, int stepGama);

  bool stabilisation(ProjectDataF * data, CalcDataF * calcData, int cntData, float stopLoopStress);

  QString ToString(bool toHtmlString = false);

  static CudaDevice * createCudaDevice(int devId = 0);

  static void printCudaDevices();

  static QString errorString(CUresult val);
  static QString attributeString(CUdevice_attribute val);
  static QString attributeString(int val);

};



#endif // CUDADEVICE_H
