#include "processor.h"

ProcessorF::ProcessorF(QString _cubinGenerator, QString _prjFileName, QString _physicsFileName) :
  calcDataCount(0), cubinGenerator(_cubinGenerator), prjFileName(_prjFileName), physicsFileName(_physicsFileName)
{
  outMolFileBase = QFileInfo(prjFileName).absolutePath() + QDir::separator() + QFileInfo(prjFileName).baseName();
  data = 0;
  calcData = 0;
  dev = 0;
  stabilisation = new Stabilisation();

  srand((unsigned int)QDateTime::currentMSecsSinceEpoch());
}

int ProcessorF::process(int devId)
{
  dev = CudaDevice::createCudaDevice(devId);
  if(dev == 0)return clearProject(EXIT_FAILURE);
  qDebug() << QObject::trUtf8("CUDA device ready [%1][%2B]%3").arg(dev->id()).arg(dev->infoMemFree()).arg(dev->name());

  QString errMsg = XMLReader::parseProjectF(&prj, prjFileName, physicsFileName);

  if(!errMsg.isEmpty()){ qDebug() << errMsg; return clearProject(EXIT_FAILURE); }
  qDebug() << QObject::trUtf8("Project [%1] readed").arg(QFileInfo(prjFileName).fileName());

  if(!checkProject())return clearProject(EXIT_FAILURE);

  if(!initData())return clearProject(EXIT_FAILURE);

  if(!data->storeMol2File(outMolFileBase, prj.name, -1, true))return clearProject(EXIT_FAILURE);

  //prepareCalcData();
  //if(!dev->loadModule("/home/basy/Projects/AtomicNetwork/cubin/proces_f.cubin"))return clearProject(EXIT_FAILURE);
  //if(!dev->stabilisation(data, calcData, calcDataCount, 0.0001f))return clearProject(EXIT_FAILURE);
  if(!stabilisation->process(&prj, data))return clearProject(EXIT_FAILURE);
  if(!data->storeMol2File(outMolFileBase, prj.name, 0, true))return clearProject(EXIT_FAILURE);


  for(int step = 1 ; step <= prj.wallMoveSteps ; step++)
  {
    applyWallMove();
    if(!stabilisation->process(&prj, data))return clearProject(EXIT_FAILURE);
    if(!data->storeMol2File(outMolFileBase, prj.name, step, true))return clearProject(EXIT_FAILURE);
    evalMoveBorderCellsX(step);
  }

  return clearProject(EXIT_SUCCESS);
}

void ProcessorF::evalMoveBorderCellsX(int step){
  float totalBorderForceSum = 0.0f, totalForceSum = 0.0f;
  for(int i = 0 ; i < data->countBorder ; i++){
    VecF curPos, matePos, dir;
    float dist, force, curForceSum = 0.0f;
    curPos.set(data->pos[i].x, data->pos[i].y, data->pos[i].z);
    for(int j = 0 ; j < MAX_MATES ; j++){
      int mateId = data->mate[i].m1[j];
      if(mateId == 0)continue;
      matePos.set(data->pos[mateId].x, data->pos[mateId].y, data->pos[mateId].z);
      if(curPos.x > matePos.x)continue;
      dir.set(matePos.x - curPos.x, matePos.y - curPos.y, matePos.z - curPos.z);
      dist = dir.length();
      force = 4.0f * prj.crystalLJE * ( pow(prj.crystalLJD, 12.0f) / pow(dist, 12.0f) - pow(prj.crystalLJD, 6.0f) / pow(dist, 6.0f) );
      dir.setLength(force);
      curForceSum += dir.x;
    }
    totalBorderForceSum += curForceSum;
  }

  for(int i = 0 ; i < data->count ; i++){
    ulong curMask = data->attr[i].mask;
    if(curMask == CELL_MASK_NULL || __IS_CELL_MASK_BUG_VAC(curMask))continue;
    VecF curPos, matePos, dir;
    float dist, force, curForceSum = 0.0f;
    curPos.set(data->pos[i].x, data->pos[i].y, data->pos[i].z);
    for(int j = 0 ; j < MAX_MATES ; j++){
      int mateId = data->mate[i].m1[j];
      if(mateId == 0)continue;
      ulong mateMask = data->attr[mateId].mask;
      matePos.set(data->pos[mateId].x, data->pos[mateId].y, data->pos[mateId].z);
      if(curPos.x > matePos.x)continue;
      dir.set(matePos.x - curPos.x, matePos.y - curPos.y, matePos.z - curPos.z);
      dist = dir.length();
      if(__IS_CELL_MASK_BUG_SUB(curMask) || __IS_CELL_MASK_BUG_SUB(mateMask)){
        force = 4.0f * prj.substitutionLJE * ( pow(prj.substitutionLJD, 12.0f) / pow(dist, 12.0f) - pow(prj.substitutionLJD, 6.0f) / pow(dist, 6.0f) );
      }else {
        force = 4.0f * prj.crystalLJE      * ( pow(prj.crystalLJD, 12.0f)      / pow(dist, 12.0f) - pow(prj.crystalLJD, 6.0f)      / pow(dist, 6.0f) );
      }
      dir.setLength(force);
      curForceSum += dir.x;
    }
    totalForceSum += curForceSum;
  }

  qDebug() << QObject::trUtf8("Step %1, Total force %2, Avg force %3 Wall force %4, Avg wall force %5")
              .arg(step)
              .arg(totalForceSum).arg(totalForceSum / (float)(data->count - data->countVac))
              .arg(totalBorderForceSum).arg(totalBorderForceSum / (float)data->countBorder);
}

void ProcessorF::applyWallMove(){
  if(prj.wallMoveType == PRESS_X){
    float sizeX = prj.wallX1 - prj.wallX0;
    float sizeXNew = prj.wallX1 - prj.wallX0 - prj.wallMoveLength;

    for(int i = 0 ; i < data->count ; i++)
    {
      float distX = prj.wallX1 - data->pos[i].x;
      float distXNew = (sizeXNew*distX)/sizeX;
      data->pos[i].x = prj.wallX1 - distXNew;
    }
  }
}


int ProcessorF::clearProject(int result)
{
  if(dev != 0){
    dev->detachContext();
    delete dev;
  }
  if(data != 0) delete data;
  if(calcData != 0)delete calcData;
  return result;
}

bool ProcessorF::checkProject()
{
  if(prj.crystalA <= 0.0f){
    qDebug() << QObject::trUtf8("Error: Invalid crystal latice parameter \"A\" !");
    return false;
  }
  if(prj.crystalLJD <= 0.0f){
    qDebug() << QObject::trUtf8("Error: Invalid Lennard-Jones potential parameter \"D\" !");
    return false;
  }
  if(prj.crystalLJE <= 0.0f){
    qDebug() << QObject::trUtf8("Error: Invalid Lennard-Jones potential parameter \"E\" !");
    return false;
  }

  if(prj.wallX0 >= prj.wallX1 || prj.wallY0 >= prj.wallY1 || prj.wallZ0 >= prj.wallZ1 ){
    qDebug() << QObject::trUtf8("Error: Invalid simulation box sizes !");
    return false;
  }

  if( prj.seed.x <= prj.wallX0 || prj.seed.x >= prj.wallX1 ||
      prj.seed.y <= prj.wallY0 || prj.seed.y >= prj.wallY1 ||
      prj.seed.z <= prj.wallZ0 || prj.seed.z >= prj.wallZ1)
  {
    prj.seed.x = (prj.wallX1 - prj.wallX0) / 2.0f;
    prj.seed.y = (prj.wallY1 - prj.wallY0) / 2.0f;
    prj.seed.z = (prj.wallZ1 - prj.wallZ0) / 2.0f;
    qDebug() << QObject::trUtf8("Warning: Seed not inside simulation box, updating to simulation box center position ...");
  }

  prj.gama.setNormal(prj.alpha, prj.beta);
  prj.beta.setNormal(prj.gama, prj.alpha);

  prj.alpha.setLength(prj.crystalA);
  prj.beta.setLength(prj.crystalA);
  prj.gama.setLength(prj.crystalA);

  return true;
}

void ProcessorF::updateABGMinMax(float x, float y, float z, VecF & min, VecF & max)
{
  float dist;
  dist = ( prj.alpha.x * x + prj.alpha.y * y + prj.alpha.z * z - prj.alpha.x * prj.seed.x - prj.alpha.y * prj.seed.y - prj.alpha.z * prj.seed.z) / prj.alpha.length();
  if(min.x > dist)min.x = dist;
  if(max.x < dist)max.x = dist;

  dist = ( prj.beta.x  * x + prj.beta.y  * y + prj.beta.z  * z - prj.beta.x  * prj.seed.x - prj.beta.y  * prj.seed.y - prj.beta.z  * prj.seed.z) / prj.beta.length();
  if(min.y > dist)min.y = dist;
  if(max.y < dist)max.y = dist;

  dist = ( prj.gama.x  * x + prj.gama.y  * y + prj.gama.z  * z - prj.gama.x  * prj.seed.x - prj.gama.y  * prj.seed.y - prj.gama.z  * prj.seed.z) / prj.gama.length();
  if(min.z > dist)min.z = dist;
  if(max.z < dist)max.z = dist;
}

bool ProcessorF::initData()
{

  if(!dev->loadModule(cubinGenerator))return false;

  VecF min(FLT_MAX), max(FLT_MIN);

  updateABGMinMax(prj.wallX0, prj.wallY0, prj.wallZ0, min, max);
  updateABGMinMax(prj.wallX0, prj.wallY0, prj.wallZ1, min, max);
  updateABGMinMax(prj.wallX0, prj.wallY1, prj.wallZ0, min, max);
  updateABGMinMax(prj.wallX0, prj.wallY1, prj.wallZ1, min, max);
  updateABGMinMax(prj.wallX1, prj.wallY0, prj.wallZ0, min, max);
  updateABGMinMax(prj.wallX1, prj.wallY0, prj.wallZ1, min, max);
  updateABGMinMax(prj.wallX1, prj.wallY1, prj.wallZ0, min, max);
  updateABGMinMax(prj.wallX1, prj.wallY1, prj.wallZ1, min, max);

  min.div(prj.crystalA);
  max.div(prj.crystalA);

  int seedIdx[3];
  seedIdx[0]  = (int)min.x - 2;
  seedIdx[1]  = (int)min.y - 2;
  seedIdx[2]  = (int)min.z - 2;

  int stepAlpha   = (int)max.x - seedIdx[0] + 2;
  int stepBeta    = (int)max.y - seedIdx[1] + 2;
  int stepGama    = (int)max.z - seedIdx[2] + 2;
  int seedOffset  = (-seedIdx[0])*stepBeta*stepGama + (-seedIdx[1])*stepGama - seedIdx[2];

  int genCellCount = stepAlpha*stepBeta*stepGama;
  if(prj.matrixType == BCC) genCellCount *= 2;
  if(prj.matrixType == FCC) genCellCount *= 4;

  VecF begin;

  begin.x = prj.seed.x + (float)seedIdx[0] * prj.alpha.x + (float)seedIdx[1] * prj.beta.x + (float)seedIdx[2] * prj.gama.x;
  begin.y = prj.seed.y + (float)seedIdx[0] * prj.alpha.y + (float)seedIdx[1] * prj.beta.y + (float)seedIdx[2] * prj.gama.y;
  begin.z = prj.seed.z + (float)seedIdx[0] * prj.alpha.z + (float)seedIdx[1] * prj.beta.z + (float)seedIdx[2] * prj.gama.z;

  //qDebug() << QObject::trUtf8("Alpha %1 ").arg(prj.alpha.toString());
  //qDebug() << QObject::trUtf8("Beta %1 ").arg(prj.beta.toString());
  //qDebug() << QObject::trUtf8("Gama %1 ").arg(prj.gama.toString());
  //qDebug() << QObject::trUtf8("Matrix(%1) begin %2 ").arg((int)prj.matrixType).arg(begin.toString());
  qDebug() << QObject::trUtf8("Generator matrix size %1 x %2 x %3 total count %4").arg(stepAlpha).arg(stepBeta).arg(stepGama).arg(genCellCount);
  //qDebug() << QObject::trUtf8("SeedIdx [%1 x %2 x %3] %4").arg(-seedIdx[0]).arg(-seedIdx[1]).arg(-seedIdx[2]).arg(seedOffset);

  size_t memSizeNeed = genCellCount * 3 * sizeof(float) + genCellCount * sizeof(long) + genCellCount * sizeof(CellMate);

  if(memSizeNeed >= dev->infoMemFree()){
    qDebug() << QObject::trUtf8("Error: Out of memory on CUDA device, required %1 free %2 !").arg(memSizeNeed).arg(dev->infoMemFree());
    dev->freeModule();
    return false;
  }

  VecF      * allPositions = new VecF[genCellCount]();
  ulong     * allMask = new ulong[genCellCount];
  CellMate  * allMate = new CellMate[genCellCount]();

  memset(allMask, 0, sizeof(ulong) * genCellCount);
  memset(allMate, 0, sizeof(CellMate) * genCellCount);

  if( !dev->prepareCellPos(prj, allPositions, begin, stepAlpha, stepBeta, stepGama) ||
      !dev->prepareCellMate(prj, allMate, stepAlpha, stepBeta, stepGama) ||
      !dev->prepareCellMask(prj, allPositions, allMask, stepAlpha, stepBeta, stepGama)){
    delete allPositions;
    delete allMask;
    delete allMate;
    dev->freeModule();
    return false;
  }

  int * newIds = new int[genCellCount];
  memset(newIds, 0, sizeof(int) * genCellCount);

  int insideCellCount = 1;
  for(int i = 0 ; i < genCellCount ; i++){
    if(allMask[i] != CELL_MASK_NULL){
      newIds[i] = insideCellCount;
      insideCellCount++;
    }
  }

  data = new ProjectDataF(insideCellCount, newIds[seedOffset]);

  insideCellCount = 1;

  int exceptMates = prj.matrixType == BCC ? 8 : 12;

  int borderCellCount = 0;

  for(int i = 0 ; i < genCellCount ; i++){
    if(allMask[i] == CELL_MASK_NULL)continue;
    data->pos[insideCellCount].set(allPositions[i].x, allPositions[i].y, allPositions[i].z);
    data->attr[insideCellCount].mask = allMask[i];
    int foundMates = 0;
    for(int j = 0 ; j < exceptMates ; j++)
    {
      int mateIdx = allMate[i].m1[j];
      if(mateIdx > 0 && mateIdx < genCellCount && newIds[mateIdx] != 0){
        CellMaskSetMate(&data->attr[insideCellCount].mask, j);
        data->mate[insideCellCount].m1[j] = newIds[mateIdx];
        foundMates++;
      }
    }

    if(exceptMates != foundMates){
      __SET_CELL_MASK_SCENE_BORDER(data->attr[insideCellCount].mask);
      borderCellCount++;
      //qDebug() << exceptMates << " " << foundMates << data->pos[insideCellCount].toString();
    }
    insideCellCount++;
  }

  int vacCout = (int)((float)data->count*prj.vacancyRatio);
  if(vacCout > 0 && vacCout < data->count/2)
  {
    for(int i = 0 ; i < vacCout ; i++){
      int randIdx = randInt(data->count);
      while(randIdx <= 0 ||
            __IS_CELL_MASK_SCENE_BORDER(data->attr[randIdx].mask) ||
            __IS_CELL_MASK_BUG_ANY(data->attr[randIdx].mask))
        randIdx = randInt(data->count);
      //__SET_CELL_MASK_BUG_VAC(data->attr[randIdx].mask);
      SetCellBug(randIdx, VACANCY, 0);
    }
    data->countVac = vacCout;
    qDebug() << QObject::trUtf8("Vacancies %1").arg(vacCout);
  }

  int subCout = (int)((float)data->count*prj.substitutionRatio);
  if(subCout > 0 && subCout < data->count/2)
  {
    for(int i = 0 ; i < subCout ; i++){
      int randIdx = randInt(data->count);
      while(randIdx <= 0 ||
            __IS_CELL_MASK_SCENE_BORDER(data->attr[randIdx].mask) ||
            __IS_CELL_MASK_BUG_ANY(data->attr[randIdx].mask))
        randIdx = randInt(data->count);
      //__SET_CELL_MASK_BUG_SUB(data->attr[randIdx].mask);
      SetCellBug(randIdx, SUBSTITUTION, 2);
    }
    qDebug() << QObject::trUtf8("Substitutions %1").arg(subCout);
  }

  for(int i = 0 ; i < data->count ; i++){
    if(!__IS_CELL_MASK_SCENE_BORDER( data->attr[i].mask))continue;
    switch(prj.wallMoveType){
    case(PRESS_X) : case(PULL_X):
      if(data->pos[i].x - prj.wallMoveLength < prj.crystalA)
        data->countBorder++;
      break;
    case(PRESS_Y) : case(PULL_Y):
      if(data->pos[i].y - prj.wallMoveLength < prj.crystalA)
        data->countBorder++;
      break;
    case(PRESS_Z) : case(PULL_Z):
      if(data->pos[i].z - prj.wallMoveLength < prj.crystalA)
        data->countBorder++;
      break;
    }
  }

  data->wallIds = new int[data->countBorder];

  int wallIdx = 0;

  for(int i = 0 ; i < data->count ; i++){
    if(!__IS_CELL_MASK_SCENE_BORDER( data->attr[i].mask))continue;
    switch(prj.wallMoveType){
    case(PRESS_X) : case(PULL_X):
      if(data->pos[i].x - prj.wallMoveLength < prj.crystalA)
        data->wallIds[wallIdx++] = i;
      break;
    case(PRESS_Y) : case(PULL_Y):
      if(data->pos[i].y - prj.wallMoveLength < prj.crystalA)
        data->wallIds[wallIdx++] = i;
      break;
    case(PRESS_Z) : case(PULL_Z):
      if(data->pos[i].z - prj.wallMoveLength < prj.crystalA)
        data->wallIds[wallIdx++] = i;
      break;
    }
  }

  qDebug() << QObject::trUtf8("Cells generated %1 (On scene border %2, on move border %3)").arg(data->count).arg(borderCellCount).arg(data->countBorder);

  return dev->freeModule();
}

void ProcessorF::SetCellBug(int id, BugType bug, int type){
  CellMaskSetBug(&data->attr[id].mask, bug);
  data->attr[id].type=type;
  for(int i = 0 ; i < MAX_MATES ; i++){
    int nearId = data->mate[id].m1[i];
    if(nearId == 0)continue;
    __SET_CELL_MASK_NEAR_BUG(data->attr[nearId].mask);
    for(int j = 0 ; j < MAX_MATES ; j++){
      int farId = data->mate[nearId].m1[j];
      if(farId == 0 || farId == id)continue;
      __SET_CELL_MASK_FAR_BUG(data->attr[farId].mask);
    }
  }
}
/*
void ProcessorF::prepareCalcData(){
  calcDataCount = 3;
  calcData = new CalcDataF[calcDataCount]();

  calcData[0].isVac = false;
  calcData[0].subType = -1;
  calcData[0].intType = -1;
  calcData[0].crStr = prj.matrixType;
  calcData[0].lA = prj.crystalLJD;
  calcData[0].lB = prj.crystalLJE;
  calcData[0].d  = prj.crystalA;
  calcData[0].aX = prj.alpha.x;
  calcData[0].aY = prj.alpha.y;
  calcData[0].aZ = prj.alpha.z;
  calcData[0].bX = prj.beta.x;
  calcData[0].bY = prj.beta.y;
  calcData[0].bZ = prj.beta.z;
  calcData[0].gX = prj.gama.x;
  calcData[0].gY = prj.gama.y;
  calcData[0].gZ = prj.gama.z;

  calcData[1].isVac = true;
  calcData[1].subType = -1;
  calcData[1].intType = -1;
  calcData[1].crStr = prj.matrixType;
  calcData[1].lA = prj.crystalLJD;
  calcData[1].lB = prj.crystalLJE;
  calcData[1].d  = prj.crystalA;
  calcData[1].aX = prj.alpha.x;
  calcData[1].aY = prj.alpha.y;
  calcData[1].aZ = prj.alpha.z;
  calcData[1].bX = prj.beta.x;
  calcData[1].bY = prj.beta.y;
  calcData[1].bZ = prj.beta.z;
  calcData[1].gX = prj.gama.x;
  calcData[1].gY = prj.gama.y;
  calcData[1].gZ = prj.gama.z;

  calcData[2].isVac = false;
  calcData[2].subType = 2;
  calcData[2].intType = -1;
  calcData[2].crStr = prj.matrixType;
  calcData[2].lA = prj.substitutionLJD;
  calcData[2].lB = prj.substitutionLJE;
  calcData[2].d  = prj.substitutionA;
  VecF tmp;
  tmp.set(prj.alpha.x, prj.alpha.y, prj.alpha.z);
  tmp.setLength(prj.substitutionA);
  calcData[2].aX = tmp.x;
  calcData[2].aY = tmp.y;
  calcData[2].aZ = tmp.z;
  tmp.set(prj.beta.x, prj.beta.y, prj.beta.z);
  tmp.setLength(prj.substitutionA);
  calcData[2].bX = tmp.x;
  calcData[2].bY = tmp.y;
  calcData[2].bZ = tmp.z;
  tmp.set(prj.gama.x, prj.gama.y, prj.gama.z);
  tmp.setLength(prj.substitutionA);
  calcData[2].gX = tmp.x;
  calcData[2].gY = tmp.y;
  calcData[2].gZ = tmp.z;
}
*/

