#include "xmlreader.h"

bool XMLReader::readRequiredI(int * out, QString * err, QDomElement & e, QStringList names)
{
  if(names.isEmpty()){
    (*err) = QObject::trUtf8("Undefined attribute names in element: <%1> !").arg(e.tagName());
    return false;
  }

  foreach(QString name, names) {
    QString val = e.attribute(name);
    if(!val.isEmpty()) {
      bool ok;
      (*out) = val.toInt(&ok);
      if(ok) return true;
      (*err) = QObject::trUtf8("Invalid value attribute %1=\"%2\" in element: <%3> !")
          .arg(name).arg(val).arg(e.tagName());
      return false;
    }
  }
  (*err) = QObject::trUtf8("Missing required attribute %1 in element: <%2> !").arg(names.first()).arg(e.tagName());
  return false;
}

bool XMLReader::readRequiredF(float * out, QString * err, QDomElement & e, QStringList names)
{
  if(names.isEmpty()){
    (*err) = QObject::trUtf8("Undefined attribute names in element: <%1> !").arg(e.tagName());
    return false;
  }

  foreach(QString name, names) {
    QString val = e.attribute(name);
    if(!val.isEmpty()) {
      bool ok;
      (*out) = val.toFloat(&ok);
      if(ok) return true;
      (*err) = QObject::trUtf8("Invalid value attribute %1=\"%2\" in element: <%3> !")
          .arg(name).arg(val).arg(e.tagName());
      return false;
    }
  }
  (*err) = QObject::trUtf8("Missing required attribute %1 in element: <%2> !").arg(names.first()).arg(e.tagName());
  return false;
}

bool XMLReader::readDefaultF(float * out, QString * err, QDomElement & e, QStringList names, float defVal)
{
  if(names.isEmpty()){
    (*err) = QObject::trUtf8("Undefined attribute names in element: <%1> !").arg(e.tagName());
    return false;
  }

  foreach(QString name, names) {
    QString val = e.attribute(name);
    if(!val.isEmpty()) {
      bool ok;
      (*out) = val.toFloat(&ok);
      if(ok) return true;
      (*err) = QObject::trUtf8("Invalid value attribute %1=\"%2\" in element: <%3> !")
          .arg(name).arg(val).arg(e.tagName());
      return false;
    }
  }
  (*out) = defVal;
  return true;
}

bool XMLReader::readCrystalMatrix(CrystalMatrixType * outType, QString * err, QDomElement & e)
{
  QString val = e.attribute("matrix");
  if(val.compare("bcc", Qt::CaseInsensitive) == 0){
    (*outType) = BCC;
    return true;
  } else if(val.compare("fcc", Qt::CaseInsensitive) == 0){
    (*outType) = FCC;
    return true;
  }
  (*err) = QObject::trUtf8("Invalid value or missing attribute matrix=\"%1\" [BCC|FCC] in element: <%2> !").arg(val).arg(e.tagName());
  return false;
}

bool XMLReader::readBorderVelocity(BorderVelocityType * outType, QString * err, QDomElement & e)
{
  QString type = e.attribute("type");
  QString dir  = e.attribute("dir");
  if(type.isEmpty() || type.compare("press", Qt::CaseInsensitive) == 0) {
    if(dir.isEmpty() || dir.compare("x", Qt::CaseInsensitive) == 0) {
      (*outType) = PRESS_X;
      return true;
    } else if(dir.compare("y", Qt::CaseInsensitive) == 0) {
      (*outType) = PRESS_Y;
      return true;
    } else if(dir.compare("z", Qt::CaseInsensitive) == 0) {
      (*outType) = PRESS_Z;
      return true;
    } else {
      (*err) = QObject::trUtf8("Invalid value attribute dir=\"%1\" [X|Y|Z] in element: <%2> !").arg(dir).arg(e.tagName());
      return false;
    }
  }
  else if (type.compare("pull", Qt::CaseInsensitive) == 0) {
    if(dir.isEmpty() || dir.compare("x", Qt::CaseInsensitive) == 0) {
      (*outType) = PULL_X;
      return true;
    } else if(dir.compare("y", Qt::CaseInsensitive) == 0) {
      (*outType) = PULL_Y;
      return true;
    } else if(dir.compare("z", Qt::CaseInsensitive) == 0) {
      (*outType) = PULL_Z;
      return true;
    } else {
      (*err) = QObject::trUtf8("Invalid value attribute dir=\"%1\" [X|Y|Z] in element: <%2> !").arg(dir).arg(e.tagName());
      return false;
    }
  }
  (*err) = QObject::trUtf8("Invalid value attribute type=\"%1\" [press|pull] in element: <%2> !").arg(type).arg(e.tagName());
  return false;
}


QString XMLReader::parseProjectF(ProjectF * prj, QString prjFileName, QString physicsFileName)
{
  QDomDocument doc("AtomicNetwork");
  QFile file(prjFileName);
  if (!file.open(QIODevice::ReadOnly))
    return QObject::trUtf8("Can't open project file !");

  QString errMsg;
  int errLine, errCol;

  if (!doc.setContent(&file, &errMsg, &errLine, &errCol)) {
      file.close();
      return QObject::trUtf8("Can't parse project file [%1][%2]:%3 !").arg(errLine).arg(errCol).arg(errMsg);
  }
  file.close();

  QDomElement rootElem = doc.documentElement();

  if(rootElem.tagName() != "AtomicNetwork")
    return QObject::trUtf8("Invalid \"AtomicNetwork\" project file !");

  prj->name = rootElem.attribute("name", "Atomic network");

  QDomElement boxElem = rootElem.firstChildElement("Box");
  if(boxElem.isNull())
    return QObject::trUtf8("Missing required element: <Box> !");

  if(!readRequiredF(&prj->wallX1, &errMsg, boxElem, (QStringList() << "x" << "X" << "x1" << "X1"))) return errMsg;
  if(!readRequiredF(&prj->wallY1, &errMsg, boxElem, (QStringList() << "y" << "Y" << "y1" << "Y1"))) return errMsg;
  if(!readRequiredF(&prj->wallZ1, &errMsg, boxElem, (QStringList() << "z" << "Z" << "z1" << "Z1"))) return errMsg;
  if(!readDefaultF( &prj->wallX0, &errMsg, boxElem, (QStringList() << "x0" << "X0"), 0.0f)) return errMsg;
  if(!readDefaultF( &prj->wallY0, &errMsg, boxElem, (QStringList() << "y0" << "Y0"), 0.0f)) return errMsg;
  if(!readDefaultF( &prj->wallZ0, &errMsg, boxElem, (QStringList() << "z0" << "Z0"), 0.0f)) return errMsg;

  QDomElement borderMoveElem = boxElem.firstChildElement("BorderVelocity");
  if(borderMoveElem.isNull())
    borderMoveElem = rootElem.firstChildElement("BorderVelocity");

  if(!borderMoveElem.isNull()) {
    if(!readBorderVelocity( &prj->wallMoveType,   &errMsg, borderMoveElem)) return errMsg;
    if(!readRequiredF(      &prj->wallMoveLength, &errMsg, borderMoveElem, (QStringList() << "stepLength" ))) return errMsg;
    if(!readRequiredI(      &prj->wallMoveSteps,  &errMsg, borderMoveElem, (QStringList() << "stepCount" ))) return errMsg;
  }

  QDomElement crystalElem = rootElem.firstChildElement("Crystal");
  if(crystalElem.isNull())
    return QObject::trUtf8("Missing required element: <Crystal> !");

  bool ok;
  QString type = crystalElem.attribute("type");
  QString temperature = crystalElem.attribute("temperature");

  if(!type.isEmpty()){
    float temperatureF = temperature.toFloat(&ok);
    if(!ok) return QObject::trUtf8("<%1> Undefined attribute \"temperature\" for type=\"%2\" !").arg(crystalElem.tagName()).arg(type);

    if(!readPhysicsCrystalF(prj, physicsFileName, type, temperatureF, &errMsg, crystalElem))return errMsg;
  }
  else{
    if(!readCrystalMatrix(&prj->matrixType, &errMsg, crystalElem)) return errMsg;
    if(!readRequiredF(    &prj->crystalA,   &errMsg, crystalElem, (QStringList() << "a" << "A"))) return errMsg;
    if(!readRequiredF(    &prj->crystalLJD, &errMsg, crystalElem, (QStringList() << "ljd" << "LJD" << "LjD"))) return errMsg;
    if(!readRequiredF(    &prj->crystalLJE, &errMsg, crystalElem, (QStringList() << "lje" << "LJE" << "LjE"))) return errMsg;
  }

  QDomElement seedElem = crystalElem.firstChildElement("Seed");
  if(!seedElem.isNull()){
    if(!readRequiredF( &prj->seed.x, &errMsg, seedElem, (QStringList() << "x" << "X"))) return errMsg;
    if(!readRequiredF( &prj->seed.y, &errMsg, seedElem, (QStringList() << "y" << "Y"))) return errMsg;
    if(!readRequiredF( &prj->seed.z, &errMsg, seedElem, (QStringList() << "z" << "Z"))) return errMsg;
  }

  QDomElement alphaElem = crystalElem.firstChildElement("Alpha");
  if(!alphaElem.isNull()){
    if(!readRequiredF( &prj->alpha.x, &errMsg, alphaElem, (QStringList() << "x" << "X"))) return errMsg;
    if(!readRequiredF( &prj->alpha.y, &errMsg, alphaElem, (QStringList() << "y" << "Y"))) return errMsg;
    if(!readRequiredF( &prj->alpha.z, &errMsg, alphaElem, (QStringList() << "z" << "Z"))) return errMsg;
  }

  QDomElement betaElem = crystalElem.firstChildElement("Beta");
  if(!betaElem.isNull()){
    if(!readRequiredF( &prj->beta.x, &errMsg, betaElem, (QStringList() << "x" << "X"))) return errMsg;
    if(!readRequiredF( &prj->beta.y, &errMsg, betaElem, (QStringList() << "y" << "Y"))) return errMsg;
    if(!readRequiredF( &prj->beta.z, &errMsg, betaElem, (QStringList() << "z" << "Z"))) return errMsg;
  }

  QDomElement gamaElem = crystalElem.firstChildElement("Gama");
  if(!gamaElem.isNull()){
    if(!readRequiredF( &prj->gama.x, &errMsg, gamaElem, (QStringList() << "x" << "X"))) return errMsg;
    if(!readRequiredF( &prj->gama.y, &errMsg, gamaElem, (QStringList() << "y" << "Y"))) return errMsg;
    if(!readRequiredF( &prj->gama.z, &errMsg, gamaElem, (QStringList() << "z" << "Z"))) return errMsg;
  }

  QDomElement vacancyElem = crystalElem.firstChildElement("Vacancy");
  if(!vacancyElem.isNull()){
    if(!readRequiredF( &prj->vacancyRatio, &errMsg, vacancyElem, (QStringList() << "ratio" ))) return errMsg;
  }

  QDomElement subElem = crystalElem.firstChildElement("Substitution");
  if(!subElem.isNull()){
    if(!readRequiredF( &prj->substitutionRatio, &errMsg, subElem, (QStringList() << "ratio" ))) return errMsg;

    QString subType = subElem.attribute("type");

    if(!subType.isEmpty()){
      float temperatureF = temperature.toFloat(&ok);
      if(!ok) return QObject::trUtf8("<%1> Undefined attribute \"temperature\" for type=\"%2\" !").arg(crystalElem.tagName()).arg(type);

      if(!readPhysicsSubstitutionF(prj, physicsFileName, type, subType, temperatureF, &errMsg, subElem))return errMsg;
    }
    else{
      if(!readRequiredF(&prj->substitutionA,   &errMsg, subElem, (QStringList() << "a" << "A")))              return errMsg;
      if(!readRequiredF(&prj->substitutionLJD, &errMsg, subElem, (QStringList() << "ljd" << "LJD" << "LjD"))) return errMsg;
      if(!readRequiredF(&prj->substitutionLJE, &errMsg, subElem, (QStringList() << "lje" << "LJE" << "LjE"))) return errMsg;
    }
  }

  return QString();
}

bool XMLReader::readPhysicsCrystalF(ProjectF * /*prj*/, QString /*physicsFileName*/, QString /*type*/, float /*temperatureF*/, QString * /*err*/, QDomElement & /*e*/)
{
  /*
  CrystalMatrixType matrixType;
  float crystalA, crystalLJE, crystalLJD;
  */
  return false;
}

bool XMLReader::readPhysicsSubstitutionF(ProjectF * /*prj*/, QString /*physicsFileName*/, QString /*type*/, QString /*subType*/, float /*temperatureF*/, QString * /*err*/, QDomElement & /*e*/)
{
  /*
  float substitutionLJD, substitutionLJE;
  */
  return false;
}
